//	TorusGames-3D-GameContents.vs
//
//	© 2014 by Jeff Weeks
//	See TermsOfUse.txt

uniform mat4	uniObjectIntoFrameCell,
				uniFrameCellIntoWorld,
				uniProjection;			//	maps world space to clip space

uniform vec4	uniColor;				//	premultiplied alpha
uniform float	uniNormalFactor;		//	If uniObjectIntoFrameCell changes normal vectors' length,
										//		uniNormalFactor corrects by the inverse factor.

uniform vec3	uniFogDistances;		//	(z_begin, z_end, 1.0/(z_end - z_begin))

#ifdef CLIP_TO_FRAME_CELL
uniform float	uniFrameCellClipDistance;	//	≈ 0.5
#endif
uniform vec4	uniObjectClipVector;	//	Defines two clip planes
										//	
										//		uniObjectClipVector · (x,y,z,1) < +1
										//		uniObjectClipVector · (x,y,z,1) > -1
										//
										//	where (x,y,z,1) is the vertex position
										//	in object coordinates.
										//	Set (0,0,0,0) to deactivate.

in vec3			atrPosition;
in vec3			atrNormal;

out vec4		varDiffuseColor;		//	premultiplied alpha

#ifdef CLIP_TO_FRAME_CELL
#ifdef GL_ES
//	To clip to the frame cell walls we need six clip distances:
//
//		x+, y+ and z+ for the walls at x|y|z == +0.5,
//	and
//		x-, y- and z- for the walls at x|y|z == -0.5.
//
//	In addition, to properly clip a tic-tac-toe win line
//	with endcaps, we may need two additional clip distances,
//	which we'll call "extra+" and "extra-".
//	For efficiency, package up all eight values as two vec4's.
out vec4		varClipDistancesA,	//	(x+, y+, z+, extra+)
				varClipDistancesB;	//	(x-, y-, z-, extra-)
#endif
#endif

//	Keep the point light source positioned at the eye point,
//	for the reason explained below, just above the call to normalize(). 
const vec3		cntLightPosition = vec3(0.0, 0.0, -1.0);


void main()
{
	vec4	tmpPositionInObjectCoords,
			tmpPositionInFrameCell;
	float	tmpObjectClipDistance;
	vec4	tmpPositionInWorld;
	float	tmpFogFactor;
	vec3	tmpLightDirection,
			tmpNormal;
	float	tmpDiffuseFactor,
			tmpBrightness;
	
	//	position

	tmpPositionInObjectCoords	= vec4(atrPosition, 1.0);
	tmpPositionInFrameCell		= uniObjectIntoFrameCell * tmpPositionInObjectCoords;


	//	clipping
	
	//	uniObjectClipVector is enabled at all times,
	//	even though it's needed only in Tic-Tac-Toe, and
	//	only for a winning three-in-a-row line with endpoints.
	//
	//	In practice, to deactivate it, set it to (0,0,0,0).
	//
	//	Note:  These two clip planes, which are always present,
	//	must be numbered 0 and 1, while the clip planes defining
	//	the faces of the fundamental cube, which are get disabled
	//	in tiling mode, must be numbered 2 through 7.  Otherwise
	//	the OpenGL driver on my Windows PC would crash spectacularly
	//	when it encounted disabled clip planes preceding enabled ones.
	//
#ifdef GL_ES	//	OpenGL ES

	//	In OpenGL ES, the code below handles uniObjectClipVector.
	//	The OpenGL ES version of 3D Torus Games supports only ViewDomain,
	//	so CLIP_TO_FRAME_CELL is always #defined in this shader. 

#else			//	desktop OpenGL 3.2 or later

	tmpObjectClipDistance = dot(uniObjectClipVector, tmpPositionInObjectCoords);
	gl_ClipDistance[0] = 1.0 - tmpObjectClipDistance;
	gl_ClipDistance[1] = tmpObjectClipDistance - (-1.0);

#endif	//	GL_ES


#ifdef CLIP_TO_FRAME_CELL

	//	Use client clip planes to trim away unwanted partial objects
	//	that extend beyond the frame cell's walls.  DrawTheScene() enables 
	//	the client clip planes for ViewDomain mode, but not for ViewTiling.

#ifdef GL_ES	//	OpenGL ES

	//	OpenGL ES doesn't provide client clip planes, so let's simulate
	//	them manually using the "discard" keyword in the fragment shader.
	//	The performance hit isn't nearly as bad as Imagination Technologies'
	//	and Apple's advice would imply.  Well, to be fair, they did say
	//	"avoid 'discard' whenever possible", not "never use it".
	//	Also, the fragment shader here is trivial, which minimizes
	//	the cost of needless fragment computations.
	//
	//	For an explanation of possible bleed-through artifacts,
	//	see the Technical Note in TorusGames-3D-GameContents.fs.
	//
	tmpObjectClipDistance = dot(uniObjectClipVector, tmpPositionInObjectCoords);
	varClipDistancesA = vec4(	1.0 - tmpObjectClipDistance,
								uniFrameCellClipDistance - vec3(tmpPositionInFrameCell));
	varClipDistancesB = vec4(	tmpObjectClipDistance - (-1.0),
								vec3(tmpPositionInFrameCell) - (-uniFrameCellClipDistance));

#else			//	desktop OpenGL 3.2 or later

	gl_ClipDistance[2] = uniFrameCellClipDistance - tmpPositionInFrameCell.x;
	gl_ClipDistance[3] = tmpPositionInFrameCell.x - (-uniFrameCellClipDistance);
	gl_ClipDistance[4] = uniFrameCellClipDistance - tmpPositionInFrameCell.y;
	gl_ClipDistance[5] = tmpPositionInFrameCell.y - (-uniFrameCellClipDistance);
	gl_ClipDistance[6] = uniFrameCellClipDistance - tmpPositionInFrameCell.z;
	gl_ClipDistance[7] = tmpPositionInFrameCell.z - (-uniFrameCellClipDistance);

#endif	//	GL_ES or desktop GL

#endif	//	CLIP_TO_FRAME_CELL


	tmpPositionInWorld = uniFrameCellIntoWorld * tmpPositionInFrameCell;

	gl_Position = uniProjection * tmpPositionInWorld;
	
	//	fog
	//
	//		tmpFogFactor = (z_end - z) / (z_end - z_begin)
	//
	//	clamped to [0,1].  (Clamping may be unnecessary.)
	//
	tmpFogFactor = clamp(
					(uniFogDistances[1] - tmpPositionInWorld[2]) * uniFogDistances[2],
					0.0,
					1.0);

	//	point light
	//
	//	Technical note:  it's possible that tmpPositionInWorld == cntLightPosition,
	//	in which case normalize() will return an undefined value,
	//	but shouldn't fail or stop the pipeline.
	//	(On my iMac's Radeon X1600, normalize(0,0,0) returns (0,0,0).)
	//	As long as cntLightPosition is the eye position,
	//	any triangle containing it will be seen edge-on,
	//	and will therefore be invisible to the user.
	//
	tmpLightDirection	= normalize(cntLightPosition - vec3(tmpPositionInWorld));
	tmpNormal			= vec3(	  uniFrameCellIntoWorld
								* uniObjectIntoFrameCell
								* vec4(atrNormal, 0.0));
	tmpDiffuseFactor	= 0.5  +  0.5 * uniNormalFactor * dot(tmpLightDirection, tmpNormal);

	//	final color
	tmpBrightness	= tmpFogFactor * tmpDiffuseFactor;
	varDiffuseColor	= vec4(tmpBrightness, tmpBrightness, tmpBrightness, 1.0) * uniColor;
}
